/***************************************************************************
 *   Copyright (C) 1998-2009 by David Bucciarelli (davibu@interfree.it)    *
 *                                                                         *
 *   This file is part of SmallptGPU.                                      *
 *                                                                         *
 *   SmallptGPU is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   SmallptGPU is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#ifndef _RENDERCONFIG_H
#define	_RENDERCONFIG_H

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <vector>

#define __CL_ENABLE_EXCEPTIONS
#define __NO_STD_VECTOR
#define __NO_STD_STRING

#if defined(__APPLE__)
#include <OpenCL/cl.hpp>
#else
#include <CL/cl.hpp>
#endif

#include <boost/thread/thread.hpp>
#include <boost/thread/barrier.hpp>

#include "camera.h"
#include "geom.h"
#include "renderdevice.h"

extern double WallClockTime();

using namespace std;

class RenderConfig {
public:
	RenderConfig(const string &sceneFileName, const unsigned int w,
		const unsigned int h, const bool useCPUs, const bool useGPUs,
		const unsigned int forceGPUWorkSize);
	~RenderConfig();

	void ReInitScene() {
		// Flush everything
		for (size_t i = 0; i < renderDevices.size(); ++i)
			renderDevices[i]->Finish();

		currentSample = 0;

		// Redownload the scene
		for (size_t i = 0; i < renderDevices.size(); ++i)
			renderDevices[i]->UpdateSceneBuffer(spheres);
	}

	void ReInit(const int reallocBuffers) {
		// Flush everything
		for (size_t i = 0; i < renderDevices.size(); ++i)
			renderDevices[i]->Finish();

		// Check if I have to reallocate buffers
		if (reallocBuffers) {
			delete pixels;
			pixels = new unsigned int[width * height];
			// Test colors
			for (size_t i = 0; i < width * height; ++i)
				pixels[i] = i;

			// Update devices
			UpdateDeviceWorkload(false);
		}

		UpdateCamera();

		currentSample = 0;
	}

	void Execute() {
		// Run the kernels
		if (currentSample < 20) {
			ExecuteKernels();
			currentSample++;
		} else {
			// After first 20 samples, continue to execute kernels for more and more time
			double startTime = WallClockTime();
			const float k = min(currentSample - 20u, 100u) / 100.f;
			const float tresholdTime = 0.5f * k;
			for (;;) {
				ExecuteKernels();
				currentSample++;

				const float elapsedTime = WallClockTime() - startTime;
				if (elapsedTime > tresholdTime)
					break;
			}
		}

		CheckDeviceWorkload();
	}

	void RestartWorkloadProcedure() {
		if (renderDevices.size() > 1) {
			for (size_t i = 0; i < renderDevices.size(); ++i)
				renderDevices[i]->ResetPerformance();

			UpdateDeviceWorkload(true);

			timeFirstWorkloadUpdate = WallClockTime();
			workLoadProfiling = true;
		}
	}

	const VECTOR_CLASS<RenderDevice *> &GetRenderDevice() const { return renderDevices; }
	const bool IsProfiling() const { return workLoadProfiling; }

	char captionBuffer[512];

	unsigned int width, height;
	unsigned int currentSample;
	unsigned int *pixels;

	Camera *camera;
	Sphere *spheres;
	unsigned int sphereCount;
	int currentSphere;

private:
	void CheckDeviceWorkload() {
		// Check if I have to update the workload
		double t = WallClockTime();

		if (workLoadProfiling && (t - timeFirstWorkloadUpdate > 10.0)) {
			UpdateDeviceWorkload(true);
			workLoadProfiling = false;
		}
	}

	void UpdateDeviceWorkload(bool calculateNewLoad) {
		cerr << "Updating OpenCL Device workloads"<< endl;

		// Define how to spread the workload
		double totalPerformance = 0.0;
		for (size_t i = 0; i < renderDevices.size(); ++i) {
			if (calculateNewLoad)
				totalPerformance += renderDevices[i]->GetPerformance();
			else
				totalPerformance += renderDevices[i]->GetPerformanceInUse();
		}

		const unsigned int totalWorkload = width * height;

		// Set the workload for each device
		unsigned int workOffset = 0;
		for (size_t i = 0; i < renderDevices.size(); ++i) {
			const unsigned int workLeft = totalWorkload - workOffset;
			unsigned int workAmount;
			if (workLeft <= 0) {
				workOffset = totalWorkload; // After the last pixel
				workAmount = 1;
			} else {
				if (i == renderDevices.size() - 1) {
					// The last device has to complete the work
					workAmount = workLeft;
				} else {
					if (calculateNewLoad)
						workAmount = workLeft * renderDevices[i]->GetPerformance() / totalPerformance;
					else
						workAmount = workLeft * renderDevices[i]->GetPerformanceInUse() / totalPerformance;
				}
			}

			renderDevices[i]->SetWorkLoad(workOffset, workAmount, width, height, pixels);

			workOffset += workAmount;
		}

		currentSample = 0;
	}

	void ExecuteKernels() {
		for (size_t i = 0; i < renderDevices.size(); ++i)
			renderDevices[i]->SetArgs(currentSample);

		// Start the rendering threads
		threadStartBarrier->wait();

		// Wait for job done signal (any reference to Bush is not itended...)
		threadEndBarrier->wait();
	}

	void UpdateCamera() {
		vsub(camera->dir, camera->target, camera->orig);
		vnorm(camera->dir);

		const Vec up = {0.f, 1.f, 0.f};
		const float fov = (FLOAT_PI / 180.f) * 45.f;
		vxcross(camera->x, camera->dir, up);
		vnorm(camera->x);
		vsmul(camera->x, width * fov / height, camera->x);

		vxcross(camera->y, camera->x, camera->dir);
		vnorm(camera->y);
		vsmul(camera->y, fov, camera->y);

		// Update devices
		for (size_t i = 0; i < renderDevices.size(); ++i)
			renderDevices[i]->UpdateCameraBuffer(camera);
	}

	void ReadScene(const char *fileName);

	void SetUpOpenCL(const bool useCPUs, const bool useGPUs,
		const unsigned int forceGPUWorkSize);

	// OpenCL fields
	VECTOR_CLASS<RenderDevice *> renderDevices;
	boost::barrier *threadStartBarrier;
	boost::barrier *threadEndBarrier;

	double timeFirstWorkloadUpdate;
	bool workLoadProfiling;
};

#endif	/* _RENDERCONFIG_H */
