/***************************************************************************
 *   Copyright (C) 1998-2009 by David Bucciarelli (davibu@interfree.it)    *
 *                                                                         *
 *   This file is part of SmallptGPU.                                      *
 *                                                                         *
 *   SmallptGPU is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   SmallptGPU is distributed in the hope that it will be useful,         *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 *                                                                         *
 ***************************************************************************/

#include <fstream>

#include "renderdevice.h"

RenderDevice::RenderDevice(const cl::Device &device, const string &kernelFileName,
		const unsigned int forceGPUWorkSize,
		const unsigned int sceneSphereCount,
		boost::barrier *startBarrier, boost::barrier *endBarrier) :
	renderThread(NULL), threadStartBarrier(startBarrier), threadEndBarrier(endBarrier),
	sphereCount(sceneSphereCount), colorBuffer(NULL), pixelBuffer(NULL),
	seedBuffer(NULL), workPerformanceInUse(1.0), exeUnitCount(0.0), exeTime(0.0) {
	deviceName = device.getInfo<CL_DEVICE_NAME > ().c_str();

	// Allocate a context with the selected device
	cl::Platform platform = device.getInfo<CL_DEVICE_PLATFORM>();
	VECTOR_CLASS<cl::Device> devices;
	devices.push_back(device);
	cl_context_properties cps[3] = {
		CL_CONTEXT_PLATFORM, (cl_context_properties)platform(), 0
	};
	context = new cl::Context(devices, cps);

	// Allocate the queue for this device
	cl_command_queue_properties prop = CL_QUEUE_PROFILING_ENABLE;
	queue = new cl::CommandQueue(*context, device, prop);

	// Create the kernel
	string src = ReadSources(kernelFileName);

	// Compile sources
	cl::Program::Sources source(1, make_pair(src.c_str(), src.length()));
	cl::Program program = cl::Program(*context, source);
	try {
		VECTOR_CLASS<cl::Device> buildDevice;
		buildDevice.push_back(device);
#if defined(__APPLE__)
		program.build(buildDevice, "-I. -D__APPLE__");
#else
		program.build(buildDevice, "-I.");
#endif
		cl::string result = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(device);
		cerr << "[Device::" << deviceName << "]" << " Compilation result: " << result.c_str() << endl;
	} catch (cl::Error err) {
		cl::string strError = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(device);
		cerr << "[Device::" << deviceName << "]" << " Compilation error:" << endl << strError.c_str() << endl;

		throw err;
	}

	kernel = new cl::Kernel(program, "RadianceGPU");

	kernel->getWorkGroupInfo<size_t>(device, CL_KERNEL_WORK_GROUP_SIZE, &workGroupSize);
	cerr << "[Device::" << deviceName << "]" << " Suggested work group size: " << workGroupSize << endl;

	// Force workgroup size if applicable and required
	if ((forceGPUWorkSize > 0) && (device.getInfo<CL_DEVICE_TYPE>() == CL_DEVICE_TYPE_GPU)) {
		workGroupSize = forceGPUWorkSize;
		cerr << "[Device::" << deviceName << "]" << " Forced work group size: " << workGroupSize << endl;
	}

	// Create the thread for the rendering
	renderThread = new boost::thread(boost::bind(RenderDevice::RenderThread, this));

	// Create camera buffer
	cameraBuffer = new cl::Buffer(*context,
				CL_MEM_READ_ONLY,
				sizeof(Camera));
	cerr << "[Device::" << deviceName << "] Camera buffer size: " << (sizeof(Camera) / 1024) << "Kb" << endl;

	sphereBuffer = new cl::Buffer(*context,
			CL_MEM_READ_ONLY,
			sizeof(Sphere) * sphereCount);
	cerr << "[Device::" << deviceName << "] Scene buffer size: " << (sizeof(Sphere) * sphereCount / 1024) << "Kb" << endl;
}

RenderDevice::~RenderDevice() {
	if (renderThread) {
		renderThread->interrupt();
		renderThread->join();

		delete renderThread;
	}

	delete queue;
	delete kernel;

	delete sphereBuffer;
	delete cameraBuffer;

	if (colorBuffer)
		delete colorBuffer;
	if (pixelBuffer)
		delete pixelBuffer;
	if (seedBuffer)
		delete seedBuffer;

	delete context;
}

void RenderDevice::RenderThread(RenderDevice *renderDevice) {
	try {
		for (;;) {
			renderDevice->threadStartBarrier->wait();

			renderDevice->SetKernelArgs();
			renderDevice->ExecuteKernel();
			renderDevice->ReadPixelBuffer();
			renderDevice->FinishExecuteKernel();
			renderDevice->Finish();

			renderDevice->threadEndBarrier->wait();
		}
	} catch (boost::thread_interrupted) {
		cerr << "[Device::" << renderDevice->GetName() << "] Rendering thread halted" << endl;
	} catch (cl::Error err) {
		cerr << "[Device::" << renderDevice->GetName() << "] ERROR: " << err.what() << "(" << err.err() << ")" << endl;
	}
}

string RenderDevice::ReadSources(const string &fileName) {
	fstream file;
	file.exceptions(ifstream::eofbit | ifstream::failbit | ifstream::badbit);
	file.open(fileName.c_str(), fstream::in | fstream::binary);

	string prog(istreambuf_iterator<char>(file), (istreambuf_iterator<char>()));
	cerr << "[Device::" << deviceName << "] Kernel file size " << prog.length() << "bytes" << endl;

	return prog;
}

void RenderDevice::SetWorkLoad(const unsigned int offset, const unsigned int amount,
		const unsigned int screenWidth,	const unsigned int screenHeght,
		unsigned int *screenPixels) {
	if (colorBuffer)
		delete colorBuffer;
	if (pixelBuffer)
		delete pixelBuffer;
	if (seedBuffer)
		delete seedBuffer;

	workOffset = offset;
	workAmount = amount;
	width = screenWidth;
	height = screenHeght;
	pixels = screenPixels;
	workPerformanceInUse = GetPerformance();

	colorBuffer = new cl::Buffer(*context,
			CL_MEM_READ_WRITE,
			sizeof(Vec) * workAmount);
	//cerr << "[Device::" << deviceName << "] Color buffer size: " << (sizeof(Vec) * workAmount / 1024) << "Kb" << endl;

	pixelBuffer = new cl::Buffer(*context,
			CL_MEM_WRITE_ONLY,
			sizeof(unsigned int) * workAmount);
	//cerr << "[Device::" << deviceName << "] Pixel buffer size: " << (sizeof(unsigned int) * workAmount / 1024) << "Kb" << endl;


	unsigned int *seeds = new unsigned int[workAmount * 2];
	for (size_t i = 0; i < workAmount * 2; i++) {
		seeds[i] = rand();
		if (seeds[i] < 2)
			seeds[i] = 2;
	}

	seedBuffer = new cl::Buffer(*context,
			CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
			sizeof(unsigned int) * workAmount * 2, seeds);
	delete seeds;
	//cerr << "[Device::" << deviceName << "] Seeds buffer size: " << (sizeof(unsigned int) * 2 * workAmount / 1024) << "Kb" << endl;

	currentSample = 0;
}
